import colorconsole, colorconsole.terminal
import sys, traceback
from optparse import OptionParser
import os, os.path
import time
import datetime

import threading

from mytools import *
from mytools.consoleadvance import *

#for cnagios
import base64
import urllib
try:
    import urllib2
except:
    import urllib.request as urllib2
import tempfile
import sys
import datetime
import BeautifulSoup
import re
 


VERSION="0.2.0"
PROG="cnagios"
DESCRIPTION="console nagios"

AUTHOR="Frederic Aoustin" 


CONF = {'color':
            {'UP' : 'GREEN',
            'DOWN' : 'RED',
            'UNREACHABLE' : 'YELLOW',
            'PENDING' : 'LGREY',
            'OK' : 'GREEN',
            'WARNING':'YELLOW',
            'UNKNOW':'PURPLE',
            'CRITICAL' : 'RED',
            'TEXT':'WHITE',
            'HEADER_BG':'GREEN',
            'HEADER_FG':'BLACK',
            'MENU_BG':'GREEN',
            'MENU_FG':'BLACK',
            },
       }

COLOR = CONF['color']       

STATUS_HOST = ["UP","DOWN","UNREACHABLE","PENDING"]
STATUS_SERVICE = ["OK","WARNING","UNKNOW","CRITICAL","PENDING"]


WINDOWS  = {
    'SERVICES' : {
        'TITLE' : 'Serv',
        'TYPE' : 'SERVICES',
        'FILTER' : {},
    },
    'HOSTS': {
        'TITLE' : 'Host',
        'TYPE' : 'HOSTS',
        'FILTER' : {},
    },
    'SERVICESNOOK' : {
        'TITLE' : 'servnoOk',
        'TYPE' : 'SERVICES',
        'FILTER' : {'valid':False},
    },
    'HOSTSNOUP': {
        'TITLE' : 'hostNoup',
        'TYPE' : 'HOSTS',
        'FILTER' : {'valid':False},
    },


}



STATUS_MAPPING = { "ack.gif" : "acknowledged",\
                   "passiveonly.gif" : "passiveonly",\
                   "disabled.gif" : "passiveonly",\
                   "ndisabled.gif" : "notifications_disabled",\
                   "downtime.gif" : "scheduled_downtime",\
                   "flapping.gif" : "flapping"}


def filtered(kwargs):
    def myfilter(x):
        try:
            test = [item for item in kwargs.keys() if x[item] == kwargs[item]]
            if len(test) == len(kwargs.keys()):
                return True
            return False    
        except:
            return False
    return myfilter    

class Callable:
    def __init__(self, anycallable):
        self.__call__ = anycallable
        
class MultipartPostHandler(urllib2.BaseHandler):
    handler_order = urllib2.HTTPHandler.handler_order - 10 # needs to run first

    def http_request(self, request):
        data = request.get_data()
        if data is not None and type(data) != str:
            v_vars = []
            try:
                for(key, value) in data.items():
                    v_vars.append((key, value))
            except TypeError:
                systype, value, traceback = sys.exc_info()
                raise TypeError("not a valid non-string sequence or mapping object").with_traceback(traceback)

            boundary, data = self.multipart_encode(v_vars)
            contenttype = 'multipart/form-data; boundary=%s' % boundary
            if(request.has_header('Content-Type')
               and request.get_header('Content-Type').find('multipart/form-data') != 0):
                print ("Replacing %s with %s" % (request.get_header('content-type'), 'multipart/form-data'))
            request.add_unredirected_header('Content-Type', contenttype)

            request.add_data(data)
        return request

    def multipart_encode(vars, boundary = None, buffer = None):
        if boundary is None:
            boundary = mimetools.choose_boundary()
        if buffer is None:
            buffer = ''
        for(key, value) in vars:
            buffer += '--%s\r\n' % boundary
            buffer += 'Content-Disposition: form-data; name="%s"' % key
            buffer += '\r\n\r\n' + value + '\r\n'
        buffer += '--%s--\r\n\r\n' % boundary
        return boundary, buffer
    
    multipart_encode = Callable(multipart_encode)
    https_request = http_request

def not_empty(x):
    '''tiny helper function for BeautifulSoup in GenericServer.py to filter text elements'''
    return bool(x.replace('&nbsp;', '').strip())


class Service(dict):

    """
    describe service by key: value, sample:
        'status': 'OK', 
        'attempt': '1/4', 
        'passiveonly': False, 
        'service': 'Total Processes', 
        'serviceid' : 'Total Processes',
        'acknowledged': False, 
        'last_check': '02-11-2014 13:57:15', 
        'last': '13:57:15',  
        'host': 'localhost', 
        'flapping': False, 
        'duration': '31d  4h 33m 32s', 
        'scheduled_downtime': False, 
        'status_information': u'PROCS OK 82 processus avec ETAT  RSZDT', 
        'info' : '', #dynamique
        'notifications_disabled': False
    """

    def __init__(self):
        super(Service, self).__init__()

    def __setitem__(self, key, value):
        if key == 'last_check':
            self['last'] = value.split(' ')[1]
        if key == 'attempt':
            self['cnt'] = value 
        if key == 'status_information':
            self['info'] = value
        if key == 'service':
            self['serviceid'] = value
            value = value[0:40]
        if key == 'status':
            self['valid'] = value == 'OK' and True or False
        super(Service, self).__setitem__(key, value)    

    def __lt__(self, other):
        if self['host'] == other['host']:
            return self['service'] < other['service']  
        return self['host'] < other['host']

    def __le__(self, other):
        if self['host'] == other['host']:
            return self['service'] <= other['service']  
        return self['host'] <= other['host']

    def __eq__(self, other):
        if self['host'] == other['host']:
            return self['service'] == other['service']  
        return self['host'] == other['host']

    def __ne__(self, other):
        if self['host'] == other['host']:
            return self['service'] != other['service']  
        return self['host'] != other['host']

    def __gt__(self, other):
        if self['host'] == other['host']:
            return self['service'] > other['service']  
        return self['host'] > other['host']

    def __ge__(self, other):
        if self['host'] == other['host']:
            return self['service'] >= other['service']  
        return self['host'] >= other['host']

class Host(dict):
    """
    describe host by key: value, sample:
        'status': 'UP',
        'attempt': 'N/A', 
        'passiveonly': False, 
        'acknowledged': False,
        'last_check': '02-11-2014 13:58:37',
        'last': '13:58:37',
        'host': 'localhost', 
        'flapping': False,
        'duration': '696d 20h 44m 56s', 
        'scheduled_downtime': False, 
        'status_information': u'PING OK   Paquets perdus  0 RTA  003 ms',
        'info' : '' #dynamique
        'notifications_disabled': False
    """

    def __init__(self):
        super(Host, self).__init__()

    def __setitem__(self, key, value):
        if key == 'last_check':
            self['last'] = value.split(' ')[1]
        if key == 'attempt':
            self['cnt'] = value 
        if key == 'status_information':
            self['info'] = value
        if key == 'status':
            self['valid'] = value == 'UP' and True or False
        super(Host, self).__setitem__(key, value)    

    def __lt__(self, other):
        return self['host'] < other['host']

    def __le__(self, other):
        return self['host'] <= other['host']

    def __eq__(self, other):
        return self['host'] == other['host']

    def __ne__(self, other):
        return self['host'] != other['host']

    def __gt__(self, other):
        return self['host'] > other['host']

    def __ge__(self, other):
        return self['host'] >= other['host']


class Nagios(object):

    def __init__(self, username, password, monitor_url, monitor_cgi_url):
        self.monitor_cgi_url = monitor_cgi_url
        self.monitor_url = monitor_url
        self.username = username
        self.password = password
        self.HTTPheaders = {}
        for giveback in ["raw", "obj"]:
            self.HTTPheaders[giveback] = {"Authorization": "Basic " + base64.b64encode(username + ":" + password)}
        self. giveback = "obj"
        self.cgi_data = None
        self.nagcgiurl_services_all =  monitor_cgi_url + "/status.cgi?host=all"
        self.nagcgiurl_hosts_all = monitor_cgi_url + "/status.cgi?hostgroup=all&style=hostdetail"

    @property
    def urlopener(self):
        # use server-owned attributes instead of redefining them with every request
        passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, self.monitor_url, self.username, self.password)
        passman.add_password(None, self.monitor_cgi_url, self.username, self.password)  
        basic_handler = urllib2.HTTPBasicAuthHandler(passman)
        digest_handler = urllib2.HTTPDigestAuthHandler(passman)
        urlopener = urllib2.build_opener(digest_handler,
                                     basic_handler,
                                     MultipartPostHandler)
        return urlopener

    def _get_services(self):
        request = urllib2.Request(self.nagcgiurl_services_all, self.cgi_data, self.HTTPheaders[self.giveback])
        urlcontent = self.urlopener.open(request)
        txt = urlcontent.read()
        parser= BeautifulSoup.BeautifulSoup(unicode(txt, "utf8", errors="ignore"))
        table = parser('table', {'class': 'status'})[0]
        if len(table('tbody')) == 0:
            trs = table('tr', recursive=False)
        else:
            tbody = table('tbody')[0]
            trs = tbody('tr', recursive=False)
        trs.pop(0)
        nagitems = []
        tds = []
        for tr in trs:
            try:
                tds = tr('td', recursive=False)
                if len(tds) > 1:
                    n = Service()#dict()
                    try:
                        n["host"] = str(tds[0](text=not_empty)[0])
                    except:
                        n["host"] = str(nagitems[len(nagitems)-1]["host"])
                    n["service"] = str(tds[1](text=not_empty)[0])
                    n["status"] = str(tds[2](text=not_empty)[0])
                    n["last_check"] = str(tds[3](text=not_empty)[0])
                    n["duration"] = str(tds[4](text=not_empty)[0])
                    n["attempt"] = str(tds[5](text=not_empty)[0]).strip()
                    if len(tds[6](text=not_empty)) == 0:
                        n["status_information"] = ""
                    else:
                        n["status_information"] = re.sub("&(#\d+|#x[0-9a-fA-F]+|\w+);",'',tds[6](text=not_empty)[0])#.encode('ascii', 'ignore')
                    n["passiveonly"] = False
                    n["notifications_disabled"] = False
                    n["flapping"] = False
                    n["acknowledged"] = False
                    n["scheduled_downtime"] = False
                    icons = tds[1].findAll('img')
                    for i in icons:
                        icon = i["src"].split("/")[-1]
                        if icon in STATUS_MAPPING:
                            n[STATUS_MAPPING[icon]] = True
                    del icons
                    nagitems.append(n)
            except Exception as e:
                #print(e)
                pass   
        return nagitems

    def _get_hosts(self):
        request = urllib2.Request(self.nagcgiurl_hosts_all, self.cgi_data, self.HTTPheaders[self.giveback])
        urlcontent = self.urlopener.open(request)
        txt = urlcontent.read()
        
        parser= BeautifulSoup.BeautifulSoup(unicode(txt, "utf8", errors="ignore"))
        table = parser('table', {'class': 'status'})[0]
        if len(table('tbody')) == 0:
            trs = table('tr', recursive=False)
        else:
            tbody = table('tbody')[0]
            trs = tbody('tr', recursive=False)
        trs.pop(0)
        tds = []
        nagitems = []
        for tr in trs:
            try:
                if len(tr('td', recursive=False)) > 1:
                    n = Host()#dict()
                    tds = tr('td', recursive=False)
                    try:
                        n["host"] = str(tds[0].table.tr.td.table.tr.td.a.string)
                    except:
                        n["host"] = str(nagitems[len(nagitems)-1]["host"])
                    n["status"] = str(tds[1].string)
                    n["last_check"] = str(tds[2].string)
                    n["duration"] = str(tds[3].string)
                    if len(tds) < 7:
                        if len(tds[4](text=not_empty)) == 0:
                            n["status_information"] = ""
                        else:
                            n["status_information"] = re.sub("&(#\d+|#x[0-9a-fA-F]+|\w+);",'', tds[4].string)#.encode("utf-8")
                        n["attempt"] = "N/A"
                    else:
                        n["attempt"] = str(tds[4].string).strip()
                        if len(tds[5](text=not_empty)) == 0:
                            n["status_information"] = ""
                        else:
                            n["status_information"] = re.sub("&(#\d+|#x[0-9a-fA-F]+|\w+);",'', tds[5].string)#.encode("utf-8")
                    n["passiveonly"] = False
                    n["notifications_disabled"] = False
                    n["flapping"] = False
                    n["acknowledged"] = False
                    n["scheduled_downtime"] = False
                    icons = tds[0].findAll('img')
                    for i in icons:
                        icon = i["src"].split("/")[-1]
                        if icon in STATUS_MAPPING:
                            n[STATUS_MAPPING[icon]] = True
                    del icons
                    nagitems.append(n)
            except Exception as e:
                #print(e)
                pass   
        return nagitems


class ThreadNagios(threading.Thread):
    def __init__(self, server, nagios, interval):
        threading.Thread.__init__(self)
        self.server = server
        self.interval = interval
        self.nagios = nagios
        self.terminated = False
    
    def stop(self):
        self.terminated = True

    def run(self):
        while not self.terminated:
            self.server.services = self.nagios._get_services()
            self.server.hosts = self.nagios._get_hosts()
            self.forcedRefresh = False
            t =0
            while not self.terminated and t < self.interval and not self.forcedRefresh:
                time.sleep(1) #if you have a problem with long time
                t =t +1
 

class ServerNagios(object):

    def __init__(self, conf, username, password, monitor_url, monitor_cgi_url):
        self.conf = conf
        self.nagios = Nagios(username, password, monitor_url, monitor_cgi_url)
        self._servers = {'nagios': ThreadNagios(self, self.nagios, self.conf.get('interval',60))}
        self._services = []
        self._hosts= []
        self._last_set = datetime.datetime.now() 

    def start(self):
        for i in self._servers:
            self._servers[i].start()

    def stop(self):
        for i in self._servers:
            self._servers[i].stop()

    @property
    def last_set(self):
        return self._last_set

    @property
    def services(self):
        return self._services

    @services.setter
    def services(self, value):
        self._services = value
        self._last_set = datetime.datetime.now()

    @property
    def hosts(self):
        return self._hosts

    @hosts.setter
    def hosts(self, value):
        self._hosts = value
        self._last_set = datetime.datetime.now()

    def get_hosts(self, **kwargs):
        if kwargs:
            return filter(filtered(kwargs), self.hosts)
        return self.hosts    

    def get_services(self, **kwargs):
        if kwargs:
            return filter(filtered(kwargs), self.services)
        return self.services    

    def refresh(self):
        self._servers['nagios'].forcedRefresh = True

def get_mask(win, width=80):
    if WINDOWS[win].get('MASQ', None):
        return WINDOWS[win]['MASQ']
    else:
        if WINDOWS[win]['TYPE'] == 'SERVICES':
            ini = "%(host)-16s|%(service)-40s|%(status)-11s|%(last)-8s|%(cnt)-3s"
            if width < 84:
                return ini
            return ini + "|%(info)s"    
        else:
            return "%(host)-16s|%(status)-11s|%(last)-10s|%(info)s"

class CNagios(Screen):

    def __init__(self, term, server = None, title=''):
        Screen.__init__(self, term)
        self.server = server
        self.title = title
        if len(self.title) > 0:
            self.set_title(self.title)
        else:
            self.set_title(PROG)
        #decalage for linux 
        if os.name == "posix":
            self.x_dec = 1       
        else:
            self.x_dec = 0
     
        self.clear()
        self.reset()
        self.no_cursor()
        self._init_screen()
        self._init_bind()

    def _init_screen(self):
        x = 3
        y = 1 + self.x_dec
        ln = int(self.width /2) -1
        self.cnt_value = {'HOSTS': {}, 'SERVICES': {}}
        Text(self, x+12, y-1, l=5, fg=COLOR['TEXT']).value='HOSTS'
        k = 0
        for i in STATUS_HOST:
            Text(self, x, y+k, l=12, fg=COLOR[i]).value = i
            self.cnt_value['HOSTS'][i] = Text(self, x+12, y+k, l=5, fg=COLOR[i])
            k = k +1
        Text(self, x+ln+9, y-1, l=8, fg=COLOR['TEXT']).value='SERVICES'
        k = 0
        for i in STATUS_SERVICE:
            Text(self, x+ln, y+k, l=9, fg=COLOR[i]).value = i
            self.cnt_value['SERVICES'][i] = Text(self, x+ln+9, y+k, l=8, fg=COLOR[i])
            k = k +1
  
        for i in self.cnt_value['SERVICES']:
            self.cnt_value['SERVICES'][i].value = '%8s' % '?'
        
        for i in self.cnt_value['HOSTS']:
            self.cnt_value['HOSTS'][i].value = '%5s' % '?'

        self.lines = []
        self.start_lines = y + max(len(STATUS_SERVICE), len(STATUS_HOST)) + 2 
        self.title_lines = Text(self, 0, self.start_lines -1, l=self.width,fg = COLOR['HEADER_FG'],bg=COLOR['HEADER_BG'])
        for i in range(0, self.height - y - max(len(STATUS_SERVICE), len(STATUS_HOST)) - 4):
                self.lines.append(Text(self,0, self.start_lines + i, l=self.width))
        subtitle = ['Quit','Refresh'] + [WINDOWS[i]['TITLE'] for i in WINDOWS.keys()]
        Menu(self,0, self.height-2, l=self.width, fg=COLOR['MENU_FG'], bg=COLOR['MENU_BG']).value = ' ' * self.x_dec + ' ' + ' '.join(subtitle)
        self.windows = 'SERVICES'
        self.pages = 0

    def _init_bind(self):
        self.bind(77,self.pageup) # ->
        self.bind(75, self.pagedown) # <-
        self.bind(67,self.pageup) # -> for linux
        self.bind(68, self.pagedown) # <- for linux
        self.bind(113,self.exit) #q
        self.bind(81, self.exit) #Q
        self.bind(114,self.refreshserver) #r
        self.bind(82, self.refreshserver) #R
        for i in WINDOWS.keys():
            char = [m for m,j in zip(WINDOWS[i]['TITLE'], WINDOWS[i]['TITLE'].upper()) if m==j and m != '']
            for k in char:
                self.bind(ord(k), self.action(i))
                self.bind(ord(k.lower()), self.action(i))
        self.start_listen()

    def pageup(self):
        self.pages = self.pages +1
        self.refresh()

    def pagedown(self):
        self.pages = self.pages > 0 and self.pages -1 or 0
        self.refresh() 

    def action(self, win):
        def action_win():
            self.windows = win
            self.pages = 0
            self.refresh()
        return action_win

    def refreshserver(self):
        self.server.refresh()

    def refresh(self):
        Text(self, self.width -8,self.height-2,l=8, fg=COLOR['MENU_FG'], bg=COLOR['MENU_BG']).value=self.server.last_set.strftime("%H:%M:%S")
        #counter
        for i in self.cnt_value['SERVICES']:
            self.cnt_value['SERVICES'][i].value = '%8s' % len(self.server.get_services(status=i))
        for i in self.cnt_value['HOSTS']:
            self.cnt_value['HOSTS'][i].value = '%5s' % len(self.server.get_hosts(status=i))
        #get lines
        if WINDOWS[self.windows]['TYPE'] == 'SERVICES':
            lin = self.server.get_services(**WINDOWS[self.windows]['FILTER'])
        else:
            lin = self.server.get_hosts(**WINDOWS[self.windows]['FILTER'])
       # print lines
        if len(lin) > 0:    
            if len(self.lines) * (self.pages + 1) > len(lin):
                self.pages = int(len(lin)/len(self.lines))
            lin = lin[len(self.lines)*self.pages:len(self.lines)*(self.pages+1)]
            masq = get_mask(self.windows, self.width)
            self.title_lines.value = masq % { i : i.replace('_',' ') for i in lin[0].keys()}
            # manage color status
            obj = {i: '' for i in lin[0].keys()}
            obj['status'] = 'X'*11
            start_status = masq % obj
            start_status = start_status.find('X') + self.x_dec
            group_by = ''
            for l, v in zip(self.lines, lin):
                txt = masq % v
                if txt.split('|')[0] == group_by:
                    txt = ' '*len(group_by) + '|' + '|'.join(txt.split('|')[1:])
                else:
                    group_by = txt.split('|')[0]    
                l.value = txt

                if start_status > -1:
                    t = Text(self, start_status, l._y, l=11, fg=COLOR[v['status']])
                    t.value = '%(status)-11s' % v
            #line blank
            for i in range(len(zip(self.lines, lin)), len(self.lines),1):
                #self.lines[i].fg = self.lines[i].dfg
                self.lines[i].value = ''
        else:
            self.title_lines.value = ' NO DATA'
            #line blank
            for i in self.lines:
                #i.fg = i.dfg
                i.value = ''
 


def load():
    usage = "usage: %prog [options]" 
    parser = OptionParser(version="%s %s" % (PROG,VERSION), usage=usage)
    parser.description= '\n'.join([ parser.get_formatwidth() % i for i in DESCRIPTION.split('\n')])
    parser.epilog = AUTHOR
    parser.add_option("-t","--title",
        dest = "title",
        help = "title of console",
        default = "")
    parser.add_option("-u","--username",
        dest = "username",
        help = "username of nagios server, default nagiosadmin",
        default = "nagiosadmin")
    parser.add_option("-p","--password",
        dest = "password",
        help = "password of username, default nagiosadmin",
        default = "nagiosadmin")
    parser.add_option("-m","--url-monitor",
        dest = "url_monitor",
        help = "url of nagios server, default http://localhost/nagios3",
        default = "http://srvpzdeb01/nagios")#"http://localhost/nagios3")
    parser.add_option("-c","--cgi-url",
        dest = "url_cgi",
        help = "url of cgi",
        default = "")
    try:
        (options, args) = parser.parse_args()
        if options.url_cgi == "":
            options.url_cgi = '%s/cgi-bin' % options.url_monitor

        server = ServerNagios({},options.username, options.password, options.url_monitor, options.url_cgi)
        server.start()
        screen = CNagios(parser.screen, server, options.title)
        if screen.width < 80:
            server.stop()
            screen.clear()
            screen.reset()
            parser.error('width of screen is < 80 columns')
            sys.exit()
        else:
            try:
                while(True):
                    screen.refresh()
                    t =0
                    while t < 10:
                        time.sleep(1) #if you have a problem with long time
                        t = t +1
            except KeyboardInterrupt as e:
                server.stop()
            screen.clear()
            screen.reset()
            sys.exit()
    except Exception as e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        formatted_lines = traceback.format_exc().splitlines()
        parser.print_stdout('\n'.join(formatted_lines),'LRED','BLACK')
        parser.error(e)
        sys.exit(1)     

